# This module implements support for TOTP 2-factor authentication, assuming
# end users use a concatenation of "password+code" as their passwords.

# The module works as follows:

# If a request is processed and there is an TOTP secret configured for the user,
# then one of the following is applied:
# - If there is a User-Password attribute in the request, the last characters
#   are assumed to be the TOTP code. If the TOTP code is good, the value
#   of User-Password is modified by removing the characters corresponding to
#   the TOTP code.
# - If there is a Cleartext-Password configured for the user, its value for this
#   request is modified by appending the expected TOTP code to it.

# In both cases, after having changed the relevant bits, the module lets
# FreeRADIUS to continue with the processing of the request.

use warnings;
use strict;
use Digest::HMAC_SHA1 qw(hmac_sha1 hmac_sha1_hex);
use Math::BigInt;
use MIME::Base32;

# Bring the global hashes into the package scope
our (%RAD_REQUEST, %RAD_CHECK);

#
# This the remapping of return values
#
use constant {
    RLM_MODULE_REJECT   => 0, # immediately reject the request
    RLM_MODULE_OK       => 2, # the module is OK, continue
    RLM_MODULE_HANDLED  => 3, # the module handled the request, so stop
    RLM_MODULE_INVALID  => 4, # the module considers the request invalid
    RLM_MODULE_USERLOCK => 5, # reject the request (user is locked out)
    RLM_MODULE_NOTFOUND => 6, # user not found
    RLM_MODULE_NOOP     => 7, # module succeeded without doing anything
    RLM_MODULE_UPDATED  => 8, # OK (pairs modified)
    RLM_MODULE_NUMCODES => 9  # How many return codes there are
};

# Same as src/include/log.h
use constant {
    L_AUTH         => 2,  # Authentication message
    L_INFO         => 3,  # Informational message
    L_ERR          => 4,  # Error message
    L_WARN         => 5,  # Warning
    L_PROXY        => 6,  # Proxy messages
    L_ACCT         => 7,  # Accounting messages
    L_DBG          => 16, # Only displayed when debugging is enabled
    L_DBG_WARN     => 17, # Warning only displayed when debugging is enabled
    L_DBG_ERR      => 18, # Error only displayed when debugging is enabled
    L_DBG_WARN_REQ => 19, # Less severe warning only displayed when debugging is enabled
    L_DBG_ERR_REQ  => 20, # Less severe error only displayed when debugging is enabled
};

### This code is copied from https://github.com/oalders/authen-oath
sub totp {
    my ( $secret, $manual_time ) = @_;
    $secret = join( "", map chr( hex($_) ), $secret =~ /(..)/g )
        if $secret =~ /^[a-fA-F0-9]{32,}$/;
    my $time = $manual_time || time();
    my $T = Math::BigInt->new( int( $time / 30 ) );
    ( my $hex = $T->as_hex ) =~ s/^0x(.*)/"0"x(16 - length $1) . $1/e;
    my $bin_code = join( "", map chr( hex($_) ), $hex =~ /(..)/g );
    my $otp = _process( $secret, $bin_code );
    return $otp;
}

sub _process {
    my ( $secret, $bin_code ) = @_;
    my $hash   = hmac_sha1($bin_code, $secret);
    my $offset = hex substr unpack( "H*" => $hash ), -1;
    my $dt     = unpack "N" => substr $hash, $offset, 4;
    $dt &= 0x7fffffff;
    $dt = Math::BigInt->new($dt);
    my $modulus = 10**6;
    return sprintf( "%06d", $dt->bmod($modulus) );
}

# Function to handle authorize
sub authorize {
    my $user_password = $RAD_REQUEST{'User-Password'};
    my $cleartext_password = $RAD_CHECK{'Cleartext-Password'};
    my $secret = $RAD_CHECK{'Moonshot-OTP-Secret'};
    if (!$secret) {
        &radiusd::radlog(L_WARN, "OTP shared secret was not found for this user");
        return RLM_MODULE_NOOP;
    }

    my $decoded_secret = MIME::Base32::decode_rfc3548($secret);
    my $expected_code = totp($decoded_secret);
    my $expected_code_prev = totp($decoded_secret, time() - 30);
    my $expected_code_next = totp($decoded_secret, time() + 30);

    if ($user_password) {
        my $otp_code = substr $user_password, -6;
        my $actual_password = substr $user_password, 0, -6;
        if ($otp_code eq $expected_code or $otp_code eq $expected_code_prev or $otp_code eq $expected_code_next){
            # rewrite password
            $RAD_REQUEST{'User-Password'} = $actual_password;
            return RLM_MODULE_OK;
        }
        else {
            &radiusd::radlog(L_ERR, "Provided OTP code ($otp_code) differs from the expected one: $expected_code");
            return RLM_MODULE_REJECT;
        }
    }
    elsif ($cleartext_password) {
        $RAD_CHECK{'Cleartext-Password'} = $cleartext_password.$expected_code;
        return RLM_MODULE_OK;
    }
    else {
        &radiusd::radlog(L_ERR, 'OTP requires config:Cleartext-Password or request:User-Password');
        return RLM_MODULE_REJECT;
    }
}

